<!-----------------------------------------------------------------------------
|
|   Onyx Portfolio & Risk Management Framework
|
|   Copyright 2014 Carlo Sbraccia
|
|   Licensed under the Apache License, Version 2.0 (the "License");
|   you may not use this file except in compliance with the License.
|   You may obtain a copy of the License at
|
|       http://www.apache.org/licenses/LICENSE-2.0
|
|   Unless required by applicable law or agreed to in writing, software
|   distributed under the License is distributed on an "AS IS" BASIS,
|   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
|   See the License for the specific language governing permissions and
|   limitations under the License.
|
------------------------------------------------------------------------------>

<div class="modal-header">
    <h3 class="modal-title">Edit Job Parameters</h3>
</div>
<div class="modal-body">
    <form novalidate class="form-horizontal" role="form">
        <div class="form-group">
            <label class="col-xs-3 control-label">Module</label>
            <div class="col-xs-8">
                <input type="text" ng-model="job.name" class="form-control"/>
            </div>
        </div>

        <div class="form-group">
            <label class="col-xs-3 control-label">Executor</label>
            <div class="col-xs-5">
                <select class="form-control"
                        ng-model="job.executor"
                        ng-options="executor for executor in executors">
                </select>
            </div>
        </div>

        <div class="form-group">
            <label class="col-xs-3 control-label">Arguments</label>
            <div class="col-xs-8">
                <input type="text" ng-model="job.kwds" class="form-control"/>
            </div>
        </div>

        <div ng-repeat="(key, value) in job.trigger">
            <div class="form-group">
                <label class="col-xs-3 control-label">{{key}}</label>
                <div class="col-xs-5">
                    <input type="text"
                           ng-model="job.trigger[key]" class="form-control"/>
                </div>
            </div>
        </div>
    </form>
</div>
<div class="modal-footer">
    <button type="button"
            class="btn btn-default btn-sm"
            ng-click="cancel()">Cancel</button>
    <button type="button"
            class="btn btn-primary btn-sm"
            ng-click="upload()">Upload Changes</button>
</div>
